/*****************************************************************************
 *
 * FILE:	util.c
 * DESCRIPTION:	Cockatrice: utilities
 * DATE:	Sat, Mar 16 2002
 * UPDATED:	Tue, Jan 15 2008
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2002-2008 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2002-2008 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: util.c,v 1.7 2008/01/15 14:07:34 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <time.h>
#if	TIME_WITH_SYS_TIME
#include <sys/time.h>
#endif	/* TIME_WITH_SYS_TIME */
#ifndef	HAVE_MKSTEMP
#include <fcntl.h>
#endif	/* HAVE_MKSTEMP */
#include <md5.h>
#include <sha.h>
#include <sha256.h>
#include <ripemd.h>
#if	ENABLE_ICONV
#include <iconv.h>
#else
#if	ENABLE_WKF
#include <wkf.h>
#endif	/* ENABLE_WKF */
#endif	/* ENABLE_ICONV */
#include "cockatrice.h"
#include "module.h"
#include "functions.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#ifndef	MAX_URI_LEN
#define	MAX_URI_LEN	(1024)
#endif	/* MAX_URI_LEN */

#ifndef	MD5_DIGEST_LENGTH
#define	MD5_DIGEST_LENGTH	(16)
#endif	/* MD5_DIGEST_LENGTH */
#ifndef	SHA_DIGEST_LENGTH
#define	SHA_DIGEST_LENGTH	(20)
#endif	/* SHA_DIGEST_LENGTH */
#ifndef	SHA256_DIGEST_LENGTH
#define	SHA256_DIGEST_LENGTH	(32)
#endif	/* SHA256_DIGEST_LENGTH */
#ifndef	RIPEMD160_DIGEST_LENGTH
#define	RIPEMD160_DIGEST_LENGTH	(20)
#endif	/* RIPEMD160_DIGEST_LENGTH */

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * FUNCTION NAME:	cc_uri_escape
 * DESCRIPTION:		encode URI by escaped characters
 * ARGUMENTS:		string
 * RETURN VALUE:	escaped string
 * Reference:		RFC 2396
 *			Uniform Resource Identifiers (URI): Generic Syntax
 */
char *
cc_uri_escape(uri)
	const char *	uri;
{
  char *	o;	/* original pointer */
  char *	s;	/* working pointer */

  if (uri == NULL) {
    cgi_errno = CGIERR_INVALID_ARGUMENTS;
    return NULL;
  }

  o = s = strdup(uri);
  if (s != NULL) {
    size_t	len;
    char *	enc;

    len = strlen(uri);
    enc = (char *)cc_calloc(len * 3 + 1, sizeof(char));
    if (enc != NULL) {
      unsigned int	i;
      unsigned int	j;

      for (i = 0, j = 0; i < len; i++) {
	if ((s[i] & 0x80) == 0x80 ||	/* Japanese EUC code */
	  /* Reserved Characters */
	  s[i] == ';' || s[i] == '/' || s[i] == '?' || s[i] == ':' ||
	  s[i] == '@' || s[i] == '&' || s[i] == '=' || s[i] == '+' ||
	  s[i] == '$' || s[i] == ',' ||
	  /* Unreserved Characters */
	  s[i] == '-' || s[i] == '_' || s[i] == '.' || s[i] == '!' ||
	  s[i] == '~' || s[i] == '*' || s[i] == '\'' || s[i] == '(' ||
	  s[i] == ')' ||
	  /* control */
	  ((char)s[i] >= 0 && (char)s[i] <= 0x1f) || (char)s[i] == 0x7f ||
	  /* space */
	  s[i] == ' ' ||
	  /* delims */
	  s[i] == '<' || s[i] == '>' || s[i] == '#' || s[i] == '%' ||
	  s[i] == '"' ||
	  /* unwise */
	  s[i] == '{' || s[i] == '}' || s[i] == '|' || s[i] == '\\' ||
	  s[i] == '^' || s[i] == '[' || s[i] == ']' || s[i] == '`'
	  ) {
	  char	b1 = (s[i] >> 4) & 0x0f;
	  char	b2 = s[i] & 0x0f;
	  enc[j++] = '%';
	  enc[j++] = b1 < 10 ? b1 + '0' : b1 - 10 + 'A';
	  enc[j++] = b2 < 10 ? b2 + '0' : b2 - 10 + 'A';
	}
	else {
	  enc[j++] = s[i];
	}
      }
      free(o);

      return (char *)cc_realloc(enc, (size_t)j);
    }
  }

  return NULL;
}

/*
 * FUNCTION NAME:	cc_uri_unescape
 * DESCRIPTION:		decode an escaped character
 * ARGUMENTS:		string
 * RETURN VALUE:	unescaped string
 * REFERENCE:		RFC 2396
 *			Uniform Resource Identifiers (URI): Generic Syntax
 */
char *
cc_uri_unescape(uri)
	const char *	uri;
{
  char *	tmp;	/* original string (to work) */
  char *	s;

  if (uri == NULL) {
    cgi_errno = CGIERR_INVALID_ARGUMENTS;
    return NULL;
  }

  tmp = s = cc_strdup(uri);
  if (tmp != NULL) {
    register char *	os  = tmp;
    register char *	ret = s;
    char		c, c1;

    while ((c = *os) != EOL) {
      /* escaped URI */
      if (c == '%') {
	os++;
	if ((c = *os++) && (c1 = *os++)) {
	  *s++ = ((c  >= 'A' ? ((c  & 0xdf) - 'A') + 10 : c  - '0') * 16) +
		 ((c1 >= 'A' ? ((c1 & 0xdf) - 'A') + 10 : c1 - '0'));
	  os--;	/* back one word */
	}
	else {
	  goto done;
	}
      }
      /* escaped white space */
      else if (c == '+') {
	*s++ = ' ';
      }
      /* normal characters */
      else {
	*s++ = c;
      }
      os++;
    }
    *s = EOL;

done:
    return ret;
  }

  return NULL;
}

/*
 * Function Name:	cc_chomp
 * Description:		remove  CRLF, CR or LF from string end
 * Arguments:		string
 * Return Value:	string
 */
char *
cc_chomp(s)
	char *	s;
{
  if (s != NULL) {
    register char *	p;
    size_t		len;

    len = strlen(s);
    p   = &s[len - 1];
    while (*p == CR || *p == LF) {
      *p-- = EOL;
    }
  }

  return s;
}

/*
 * FUNCTION NAME:	cc_digest_string
 * DESCRIPTION:		digest string
 * ARGUMENTS:		string, digest_t
 * RETURN VALUE:	Successful: Digest, otherwise: NULL
 */
char *
cc_digest_string(str, type)
	const char *	str;
	digest_t	type;
{
  size_t	len;

  if (str == NULL) {
    cgi_errno = CGIERR_INVALID_ARGUMENTS;
    return NULL;
  }

  len = strlen(str);
  switch (type) {
    case CC_DIGEST_MD5: {
	MD5_CTX			md5_ctx;
	static unsigned char	md5_hex[MD5_DIGEST_LENGTH * 2 + 1];

	memset(md5_hex, 0, sizeof(md5_hex));
	MD5Init(&md5_ctx);
	MD5Update(&md5_ctx, str, len);
	MD5End(&md5_ctx, md5_hex);

	return md5_hex;
      }
      break;
    case CC_DIGEST_SHA1: {
	SHA_CTX			sha_ctx;
	static unsigned char	sha_hex[SHA_DIGEST_LENGTH * 2 + 1];

	memset(sha_hex, 0, sizeof(sha_hex));
	SHA1_Init(&sha_ctx);
	SHA1_Update(&sha_ctx, str, len);
	SHA1_End(&sha_ctx, sha_hex);

	return sha_hex;
      }
      break;
    case CC_DIGEST_SHA256: {
	SHA256_CTX		sha256_ctx;
	static unsigned char	sha256_hex[SHA256_DIGEST_LENGTH * 2 + 1];

	memset(sha256_hex, 0, sizeof(sha256_hex));
	SHA256_Init(&sha256_ctx);
	SHA256_Update(&sha256_ctx, str, len);
	SHA256_End(&sha256_ctx, sha256_hex);

	return sha256_hex;
      }
      break;
    case CC_DIGEST_RMD160: {
	RIPEMD160_CTX		rmd160_ctx;
	static unsigned char	rmd160_hex[RIPEMD160_DIGEST_LENGTH * 2 + 1];

	memset(rmd160_hex, 0, sizeof(rmd160_hex));
	RIPEMD160_Init(&rmd160_ctx);
	RIPEMD160_Update(&rmd160_ctx, str, len);
	RIPEMD160_End(&rmd160_ctx, rmd160_hex);

	return rmd160_hex;
      }
      break;
    default:
      break;
  }

  return NULL;
}

#if	ENABLE_ICONV
/* This order is important */
static const char *	cc_iconv_guess[] = {
  "UTF-8",
  "EUCJP-MS", "EUC-JP",
  "CP932", "SHIFT_JIS",
  "ISO-2022-JP-2", "ISO-2022-JP-1", "ISO-2022-JP",
  "UCS-2BE", "UCS-2LE", "UCS-2", "UTF-16BE", "UTF-16LE", "UTF-16",
  "UCS-4BE", "UCS-4LE", "UCS-4", "UTF-32BE", "UTF-32LE", "UTF-32",
};

/* only japanese character set */
static int
cc_conv_any2xxx(inbuf, insize, outcode, outbuf, outsize)
	const char *	inbuf;
	size_t		insize;
	const char *	outcode;
	char *		outbuf;
	size_t		outsize;
{
  register size_t	i;

  for (i = 0; i < sizeof(cc_iconv_guess) / sizeof(const char *); i++) {
    iconv_t	ic;

    ic = iconv_open(outcode, cc_iconv_guess[i]);
    if (ic != (iconv_t)-1) {
      const char *	ptr_in	= inbuf;
      char *		ptr_out	= outbuf;
      size_t		len_in  = insize;
      size_t		len_out = outsize;
      size_t		rc;

      memset(outbuf, 0, outsize);
      rc = iconv(ic, &ptr_in, &len_in, &ptr_out, &len_out);
      iconv_close(ic);
      if (rc != (size_t)-1) {
#if	DEBUG
	fprintf(stderr, "DEBUG[guess] %s(%zu) -> %s(%zu) : %s\n",
		cc_iconv_guess[i], insize, outcode, len_out, inbuf);
#endif	/* DEBUG */
	return 0;
      }
    }
  }

  return -1;
}

int
cc_conv_any2euc(inbuf, insize, outbuf, outsize)
	const char *	inbuf;
	size_t		insize;
	char *		outbuf;
	size_t		outsize;
{
  const char *	guess[] = { "ISO-2022-JP",  "UTF-8", "UTF-16", "UTF-32", "SHIFT_JIS", "EUC-JP" };
  register size_t	i;

  for (i = 0; i < sizeof(guess) / sizeof(const char *); i++) {
    iconv_t	ic;

    ic = iconv_open("EUC-JP", guess[i]);
    if (ic != (iconv_t)-1) {
      const char *	ptr_in	= inbuf;
      char *		ptr_out	= outbuf;
      size_t		len_in  = insize;
      size_t		len_out = outsize;
      size_t		rc;

      memset(outbuf, 0, outsize);
      rc = iconv(ic, &ptr_in, &len_in, &ptr_out, &len_out);
      iconv_close(ic);
      if (rc != (size_t)-1) {
	return 0;
      }
    }
  }

  return -1;
}
#else
#if	ENABLE_WKF
static int
cc_conv_any2xxx(inbuf, insize, outcode, outbuf, outsize)
	const char *	inbuf;
	size_t		insize;
	const char *	outcode;
	char *		outbuf;
	size_t		outsize;
{
  cgi_errno = CGIERR_NOT_IMPLEMENTED;

  return -1;
}

int
cc_conv_any2euc(inbuf, insize, outbuf, outsize)
	const char *	inbuf;
	size_t		insize;
	char *		outbuf;
	size_t		outsize;
{
  conv_t	rc;

  wkfSetLineEndCode(LE_LF);

  rc = wkfConvertKanjiCodeOfString(KC_UNKNOWN, inbuf, KC_EUC, outbuf, outsize);

  return (rc == CONV_ERR) ? -1 : 0;
}
#endif	/* ENABLE_WKF */
#endif	/* ENABLE_ICONV */

/*****************************************************************************/

#ifndef	HAVE_MKSTEMP
/*
 * FUNCTION NAME:	mkstemp
 * DESCRIPTION:		create temporary file and return its file descriptor
 * ARGUMENTS:		template path
 * RETURN VALUE:	file descriptor
 */
int
mkstemp(path)
	char *	path;
{
  char *	p = mktemp(path);

  if (p == NULL) {
    return -1;
  }

  return open(path, O_CREAT|O_EXCL|O_RDWR, 0600);
}
#endif	/* HAVE_MKSTEMP */

/*****************************************************************************/

static cmURI	uri_funcs = {
  cc_uri_escape,
  cc_uri_unescape
};

static cmUtil	util_funcs = {
  cc_chomp,
  cc_digest_string,
  cc_conv_any2xxx,
  cc_conv_any2euc
};

void
cc_utils_register_hooks(void)
{
  cc_hook_module_uri(&uri_funcs);
  cc_hook_module_util(&util_funcs);
}
