/*****************************************************************************
 *
 * FILE:	mod_file.c
 * DESCRIPTION:	Cockatrice: file lock utilities (module)
 * DATE:	Sat, Mar 16 2002
 * UPDATED:	Fri, Jul  9 2004
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2002-2004 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2002-2004 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_file.c,v 1.1 2007/08/19 16:49:15 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#include "cockatrice.h"
#include "module.h"
#include "functions.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
typedef struct _LockManager {
  char *		name;
  char *		path;
  struct _LockManager *	next;
} LockManager;

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void		cm_file_finish(void);
static int		cm_file_lock(const char *);
static int		cm_file_unlock(const char *);
static const char *	basename(const char *);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static LockManager *	locks = NULL;

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * FUNCTION NAME:	cm_file_finish
 * DESCRIPTION:		unlock all locked files
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
static void
cm_file_finish(void)
{
  register LockManager *	lmp;
  register LockManager *	lmp_next;

  for (lmp = locks; lmp != NULL; lmp = lmp_next) {
    lmp_next = lmp->next;
    if (lmp->path != NULL) {
      unlink(lmp->path);
    }
    cc_free(lmp->name);
    cc_free(lmp->path);
    free(lmp);
  }
}

/*
 * FUNCTION NAME:	cm_file_lock
 * DESCRIPTION:		lock file exclusively
 * ARGUMENTS:		filename
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
cm_file_lock(filename)
	const char *	filename;
{
  LockManager *	lmp;
  int		wait = 5;	/* max waiting time */

  if (filename == NULL) {
    cgi_errno = CGIERR_INVALID_ARGUMENTS;
    return -1;
  }

  lmp = (LockManager *)cc_calloc(1, sizeof(LockManager));
  if (lmp == NULL) {
    return -1;
  }

  if (asprintf(&lmp->path, "%s.lock", basename(filename)) == -1) {
    cgi_errno = CGIERR_NO_ENOUGH_MEMORY;
    free(lmp);
    return -1;
  }
  lmp->name = cc_strdup(filename);
  if (lmp->name == NULL) {
    free(lmp->path);
    free(lmp);
    return -1;
  }

  /* lock the file */
  while (symlink(".", lmp->path) == -1) {
    if (wait-- < 0) {
      free(lmp->name);
      free(lmp->path);
      free(lmp);
      cgi_errno = CGIERR_OTHER_PROCESS_USING;
      return -1;
    }
#if	HAVE_USLEEP
    usleep(200000);	/* 0.2 second */
#else
    sleep(1);
#endif	/* HAVE_USLEEP */
  }
  lmp->next	= locks;
  locks		= lmp;

  return 0;
}

/*
 * FUNCTION NAME:	cm_file_unlock
 * DESCRIPTION:		unlock file
 * ARGUMENTS:		filename
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
cm_file_unlock(filename)
	const char *	filename;
{
  register LockManager *	lmp;

  for (lmp = locks; lmp != NULL; lmp = lmp->next) {
    if (filename != NULL && strcmp(lmp->name, filename) == 0) {
      if (unlink(lmp->path) != -1) {
	cc_free(lmp->name);
	cc_free(lmp->path);
	return 0;
      }
    }
  }

  return -1;
}

/*
 * Function Name:	basename
 * Description:		return filename portion of pathname
 * Arguments:		file path
 * Return Value:	Successful: basename, otherwise: NULL
 */
static const char *
basename(path)
	const char *	path;
{
  char *	p;

  if (path == NULL) {
    return NULL;
  }

  p = (char *)strrchr(path, '/');
  if (p == NULL) {
    return path;
  }
  return ++p;
}

/*****************************************************************************/

static const cmFile	file_funcs = {
  cm_file_lock,
  cm_file_unlock
};

static void
cm_file_register_hooks(void)
{
  cc_hook_module_file(&file_funcs);
}

static ccModule	mod_file = {
  CC_MODULE_MAGIC_NUMBER_MAJOR,		/* API version */
  "mod_file",				/* module name */
  "Cockatrice: file module 1.0",	/* description such as version etc.. */
  NULL,					/* module handle */
  (void *)NULL,				/* called when it's loaded */
  (void *)cm_file_finish,		/* called when it's unloaded */
  (void *)cm_file_register_hooks
};

ccModule *
ccModuleInfo(void)
{
  return &mod_file;
}
