/*****************************************************************************
 *
 * FILE:	mod_sqlite3.c
 * DESCRIPTION:	Cockatrice: SQLite3 database operator (module)
 * DATE:	Mon, May  3 2004
 * UPDATED:	Wed, Jan 23 2008
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2004-2008 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2004-2008 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_sqlite3.c,v 1.4 2008/01/30 05:54:54 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <sqlite3.h>
#include "cockatrice.h"
#include "module.h"
#include "ccpath.h"
#include "functions.h"

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	cm_sqlite3_finish(void);
static int	cm_sqlite3_open(const char *);
static int	cm_sqlite3_close(void);
static int	cm_sqlite3_exec(const char *);
static int	cm_sqlite3_query(const char *);
static int	cm_sqlite3_fetch(char ***);
static int	cm_sqlite3_fetch_done(void);
static int	cm_sqlite3_fetch_reset(void);
static char *	cm_sqlite3_quote(const char *);
static char *	get_sqlite_db_path(const char *);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static sqlite3 *	sdb = NULL;
static sqlite3_stmt *	svm = NULL;

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

static void
cm_sqlite3_finish(void)
{
  cm_sqlite3_close();
}

/*
 * FUNCTION NAME:	cm_sqlite3_open
 * DESCRIPTION:		open database
 * ARGUMENTS:		database name
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
cm_sqlite3_open(dbname)
	const char *	dbname;
{
  char *	path;
  int		rc;

  path = get_sqlite_db_path(dbname);
  if (path == NULL) {
    cgi_errno = CGIERR_DATABASE_OPEN_FAILURE;
    return -1;
  }

  if (sdb != NULL) {
    sqlite3_close(sdb);
  }
  rc = sqlite3_open(path, &sdb);	free(path);
  if (rc != SQLITE_OK) {
    fprintf(stderr, "Failed to open database: %s\n", sqlite3_errmsg(sdb));
    cgi_errno = CGIERR_DATABASE_OPEN_FAILURE;
    return -1;
  }

  return 0;
}

/*
 * FUNCTION NAME:	cm_sqlite3_close
 * DESCRIPTION:		close database
 * ARGUMENTS:		none
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
cm_sqlite3_close(void)
{
  if (svm != NULL) {
    sqlite3_finalize(svm);
    svm = NULL;
  }
  if (sdb != NULL) {
    sqlite3_close(sdb);
    sdb = NULL;
  }

  return 0;
}

/*
 * FUNCTION NAME:	cm_sqlite3_exec
 * DESCRIPTION:		exec a sql statement
 * ARGUMENTS:		SQL statement
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
cm_sqlite3_exec(sql)
	const char *	sql;
{
  char *	errMsg;
  int		rc;

  if (sql == NULL) { return -1; }	/* no SQL query */
  if (sdb == NULL) { return -1; }	/* no opened database */
  if (svm != NULL) { return -1; }	/* called after cm_sqlite_query() */
  rc = sqlite3_exec(sdb, sql, NULL, NULL, &errMsg);
  if (rc != SQLITE_OK) {
    fprintf(stderr, "Failed to execute a SQL: %s\n", errMsg);
    sqlite3_free(errMsg);
    return -1;
  }

  return 0;
}

/*
 * FUNCTION NAME:	cm_sqlite3_query
 * DESCRIPTION:		send a SQL query
 * ARGUMENTS:		SQL statement
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
cm_sqlite3_query(sql)
	const char *	sql;
{
  const char *	tail;
  int		rc;

  if (sql == NULL) { return -1; }	/* no SQL query */
  if (sdb == NULL) { return -1; }	/* no opened database */
  if (svm != NULL) {
    sqlite3_finalize(svm);
    svm = NULL;
  }
  rc = sqlite3_prepare_v2(sdb, sql, (int)strlen(sql), &svm, &tail);
  if (rc != SQLITE_OK) {
    fprintf(stderr, "Failed to send a query: %s\n", sqlite3_errmsg(sdb));
    return -1;
  }

  return 0;
}

/*
 * FUNCTION NAME:	cm_sqlite3_fetch
 * DESCRIPTION:		fetch a result
 * ARGUMENTS:		pointer for result
 * RETURN VALUE:	done: 0, error: -1, column of result: >0
 */
static int
cm_sqlite3_fetch(retval)
	char ***	retval;
{
  if (sdb == NULL) { return -1; }	/* no opened database */
  if (svm != NULL) {
    switch (sqlite3_step(svm)) {
      case SQLITE_ROW: {
	  int	n;

	  n = sqlite3_column_count(svm);
	  if (n > 0) {
	    char **	value;

	    value = (char **)cc_calloc((size_t)n, sizeof(char *));
	    if (value != NULL) {
	      register int	i;

	      for (i = 0; i < n; i++) {
		const unsigned char *	v;

		v = sqlite3_column_text(svm, i);
		if (v != NULL) {
		  value[i] = (char *)cc_strdup(v);
		}
	      }
	      if (retval != NULL) {
		*retval = value;
		return n;
	      }
	    }
	  }
	}
      case SQLITE_DONE:
	sqlite3_finalize(svm);	svm = NULL;
	return 0;
      case SQLITE_BUSY:
      case SQLITE_MISUSE:
      case SQLITE_ERROR:
      default:
	break;
    }
  }

  return -1;
}

static int
cm_sqlite3_fetch_done(void)
{
  if (sdb != NULL) {
    if (svm != NULL) {
      sqlite3_finalize(svm);
      svm = NULL;
      return 0;
    }
  }
  return -1;
}

static int
cm_sqlite3_fetch_reset(void)
{
  if (sdb != NULL) {
    if (svm != NULL) {
      sqlite3_reset(svm);
      return 0;
    }
  }
  return -1;
}

/*
 * FUNCTION NAME:	cm_sqlite3_quote
 * DESCRIPTION:		quote a SQL statement
 * ARGUMENTS:		SQL statement
 * RETURN VALUE:	successful: a quoted SQL statement, otherwise: NULL
 */
static char *
cm_sqlite3_quote(sql)
	const char *	sql;
{
  char *	retval = NULL;

  if (sql != NULL) {
    size_t		n;	/* total number of single quote */
    size_t		l;	/* length of sql */
    const char *	c;

    for (n = l = 0, c = sql; *c != '\0'; l++, c++) {
      if (*c == '\'') { n++; /* count up */ }
    }
    if (n > 0) {	/* includes single quote */
      char *	q;

      retval = q = (char *)calloc(l + n + 1, sizeof(char));
      if (retval) {

	for (c = sql; *c != '\0'; c++) {
	  switch (*c) {
	    case '\'': *q++ = '\''; *q++ = '\''; break;
	    default:   *q++ = *c; break;
	  }
	}
	*q++ = '\0';
      }
    }
    else {
      retval = strdup(sql);
    }
  }

  return retval;
}

/*
 * FUNCTION NAME:	get_sqlite_db_path
 * DESCRIPTION:		return sqlite database path
 * ARGUMENTS:		database name
 * RETURN VALUE:	Successful: path, otherwise: NULL
 */
static char *
get_sqlite_db_path(dbname)
	const char *	dbname;
{
  char *	path;
  int		n;

  /* error checker */
  if (dbname == NULL) { return NULL; }

  /* absolete path */
  if (dbname[0] == '/') { return cc_strdup(dbname); }

  n = asprintf(&path, "%s/%s.%s", SQLITE_DB_DIR, dbname, SQLITE_DB_SUFFIX);

  return n == -1 ? NULL : path;
}

/*****************************************************************************/

static const cmDatabase	sqlite3_funcs = {
  cm_sqlite3_open,
  cm_sqlite3_close,
  cm_sqlite3_exec,
  cm_sqlite3_query,
  cm_sqlite3_fetch,
  cm_sqlite3_fetch_done,
  cm_sqlite3_fetch_reset,
  cm_sqlite3_quote
};

static void
cm_sqlite3_register_hooks(void)
{
  cc_hook_module_database(&sqlite3_funcs);
}

static ccModule	mod_sqlite3 = {
  CC_MODULE_MAGIC_NUMBER_MAJOR,		/* API version */
  "mod_sqlite3",			/* module name */
  "Cockatrice: sqlite3 Module 1.0.3",	/* description such as version etc.. */
  NULL,					/* module handle */
  (void *)NULL,				/* called when it's loaded */
  (void *)cm_sqlite3_finish,		/* called when it's unloaded */
  (void *)cm_sqlite3_register_hooks
};

ccModule *
ccModuleInfo(void)
{
  return &mod_sqlite3;
}
